/********************************************************************************
 *   dciTask -- compact task manager for PocketPC.                              *
 *                                                                              * 
 *   Copyright (C) 2004-2007 dci (dev@cpp.in)                                   *
 *                                                                              *
 *   This program is free software; you can redistribute it and/or              *
 *   modify it under the terms of the GNU General Public License version 2      *
 *   as published by the Free Software Foundation.                              *
 *                                                                              *
 *   This program is distributed in the hope that it will be useful,            *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU          *
 *   General Public License version 2 for more details.                         *
 *                                                                              *
 *   You should have received a copy of the GNU General Public License          *
 *   version 2 along with this library; if not, write to the Free Software      *
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA  *
 ********************************************************************************
 */


#include <windows.h>



#define DCISIG 0x696364 /* dci\0x00 */

typedef struct wndlist_tag
{
	HWND      hWnd;
	HICON     hIcon;
	BOOL      bStockIcon;
	wchar_t * sText;
}
wndlist_t;



LONG   lOrigProc;
HFONT  hMenuFont;

wndlist_t * wndlist = NULL;
unsigned wndlistcnt = 0;
unsigned wndlistmax = 0;


/* hooking defs */
#define WH_KEYBOARD_LL (20)
typedef LRESULT ( CALLBACK * HOOKPROC ) ( int nCode, WPARAM wParam, LPARAM lParam);
typedef HHOOK   ( WINAPI   * SETWHOOK ) ( int idHook, HOOKPROC lpfn, HINSTANCE hmod, DWORD dwThreadId );
typedef BOOL    ( WINAPI   * UNHOOKWH ) ( HHOOK hhk );
typedef LRESULT ( WINAPI   * CALLNEXT ) ( HHOOK hhk, int nCode, WPARAM wParam, LPARAM lParam );
typedef struct {
    DWORD vkCode;
    DWORD scanCode;
    DWORD flags;
    DWORD time;
    ULONG_PTR dwExtraInfo;
} KBDLLHOOKSTRUCT, *PKBDLLHOOKSTRUCT;

/* hooking globals */
HHOOK g_hHook;
BOOL  g_bHookRightPressed;
UINT  g_iHookDigitPressed;
SETWHOOK fpSetWindowsHookEx;
UNHOOKWH fpUnhookWindowsHookEx;
CALLNEXT fpCallNextHookEx;

/* hooking callback */
LRESULT HookProc ( int nCode, WPARAM wParam, LPARAM lParam )
{
	if ( nCode >= 0 )
	{
		LRESULT ret = 0;
		KBDLLHOOKSTRUCT * l_info = ( KBDLLHOOKSTRUCT * ) lParam;
		/* */
		if ( l_info->vkCode == VK_RIGHT )
		{
			g_bHookRightPressed = TRUE;
			ret = 1;
		}
		else if ( ( l_info->vkCode >= 0x31 ) && ( l_info->vkCode <= 0x39 ) )
		{
			g_iHookDigitPressed = l_info->vkCode - 0x30;
			ret = 1;
		}
		else if ( l_info->vkCode == 0x30 )
		{
			g_iHookDigitPressed = 10;
			ret = 1;
		}
		/* send enter only if key released */
		if ( ret && ( 0 == ( GetAsyncKeyState ( l_info->vkCode ) &~ 1 ) ) )
		{
			keybd_event ( VK_RETURN, 0, KEYEVENTF_SILENT, 0 );
			keybd_event ( VK_RETURN, 0, KEYEVENTF_SILENT | KEYEVENTF_KEYUP, 0 );
		}
		/* */
		return ret;
	}
	return fpCallNextHookEx ( g_hHook, nCode, wParam, lParam );
}

BOOL CanCloseWindow ( HWND hWnd )
{
	DWORD dwShell, dwTest;
	GetWindowThreadProcessId ( GetDesktopWindow ( ), &dwShell );
	GetWindowThreadProcessId ( hWnd, &dwTest );
	return ( dwTest == dwShell ) ? FALSE : TRUE;
}

LRESULT FilterProc ( HWND hWnd, UINT nMsg, WPARAM wParam, LPARAM lParam )
{
	if ( ( nMsg == WM_LBUTTONDOWN ) || ( nMsg == WM_LBUTTONUP ) )
	{
		RECT rtBar;
		GetClientRect ( hWnd, &rtBar );
		if ( /* xPos */ LOWORD(lParam) > ( rtBar.right - ( rtBar.bottom / 3 ) ) )
		{
			if ( nMsg == WM_LBUTTONUP )
			{
				HWND hIterateWnd;
				LOGFONT lfMenuFont;
				HKEY hFontInfoKey;
				DWORD dummy;
				HMENU hPopup;
				DWORD imenu;
				/* populate windows list */
				wndlistcnt = 0;
				wndlistmax = 64;
				wndlist = ( wndlist_t * ) malloc ( wndlistmax * sizeof ( wndlist_t ) );
				/* */
				hIterateWnd = GetWindow ( GetDesktopWindow ( ), GW_HWNDFIRST );
				while ( hIterateWnd )
				{
					if ( IsWindowVisible ( hIterateWnd ) )
					{
						unsigned long wtextlen = GetWindowTextLength ( hIterateWnd );
						if ( wtextlen )
						{
							wchar_t * wtext = ( wchar_t * ) malloc ( sizeof ( wchar_t ) * ( wtextlen + 1 ) );
							GetWindowText ( hIterateWnd, wtext, wtextlen + 1 );
							if ( _wcsicmp ( wtext, L"Phone" ) && _wcsicmp ( wtext, L"MS_SIPBUTTON" ) && _wcsicmp ( wtext, L"CursorWindow" ) && _wcsicmp ( wtext, L"Pocket PC Emulation" ) )
							{
								/* */
								wndlist[wndlistcnt].hWnd = hIterateWnd;
								/* get icon */
								wndlist[wndlistcnt].hIcon = ( HICON ) SendMessage ( hIterateWnd, WM_GETICON, ICON_SMALL, 0 );
								wndlist[wndlistcnt].bStockIcon = TRUE;
								if ( NULL == wndlist[wndlistcnt].hIcon )
								{
									DWORD   pid;
									wchar_t mpath[MAX_PATH];
									GetWindowThreadProcessId ( hIterateWnd, &pid );
									GetModuleFileName ( ( HMODULE ) pid, mpath, MAX_PATH );
									ExtractIconEx ( mpath, 0, NULL, &wndlist[wndlistcnt].hIcon, 1 );
									wndlist[wndlistcnt].bStockIcon = FALSE;
								}
								/* */
								wndlist[wndlistcnt].sText = wtext;
								/* realloc? */
								if ( ++wndlistcnt == wndlistmax )
								{
									wndlistmax *= 2;
									wndlist = ( wndlist_t * ) realloc ( wndlist, wndlistmax * sizeof ( wndlist_t ) );
								}
							}
							else
							{
								free ( wtext );
							}
						}
					}
					/* */
					hIterateWnd = GetWindow ( hIterateWnd, GW_HWNDNEXT );
				}
				/* create menu font */
				hMenuFont = NULL;
				memset ( &lfMenuFont, 0, sizeof ( lfMenuFont ) );
				if ( ERROR_SUCCESS == RegCreateKeyEx ( HKEY_LOCAL_MACHINE, L"System\\GWE\\Menu\\PopFnt", 0, NULL, 0, ( REGSAM ) NULL, NULL, &hFontInfoKey, &dummy ) )
				{
					DWORD fsize;
					DWORD dwHtInPts;
					/* */
					fsize = sizeof ( lfMenuFont.lfHeight );
					RegQueryValueEx ( hFontInfoKey, L"Ht", NULL, &dummy, ( BYTE * ) &lfMenuFont.lfHeight, &fsize );
					dwHtInPts = 0;
					fsize = sizeof ( dwHtInPts );
					RegQueryValueEx ( hFontInfoKey, L"HtInPts", NULL, &dummy, ( BYTE * ) &dwHtInPts, &fsize );
					if ( dwHtInPts )
					{
						HDC tDC = GetDC ( hWnd );
						lfMenuFont.lfHeight = - ( ( GetDeviceCaps ( tDC, LOGPIXELSY ) * ( lfMenuFont.lfHeight + 96 ) ) / 7200 );
						ReleaseDC ( hWnd, tDC );
					}
					/* */
					fsize = sizeof ( lfMenuFont.lfWeight );
					RegQueryValueEx ( hFontInfoKey, L"Wt", NULL, &dummy, ( BYTE * ) &lfMenuFont.lfWeight, &fsize );
					fsize = sizeof ( lfMenuFont.lfCharSet );
					RegQueryValueEx ( hFontInfoKey, L"CS", NULL, &dummy, ( BYTE * ) &lfMenuFont.lfCharSet, &fsize );
					fsize = sizeof ( lfMenuFont.lfItalic );
					RegQueryValueEx ( hFontInfoKey, L"It", NULL, &dummy, ( BYTE * ) &lfMenuFont.lfItalic, &fsize );
					fsize = sizeof ( lfMenuFont.lfFaceName );
					RegQueryValueEx ( hFontInfoKey, L"Nm", NULL, &dummy, ( BYTE * ) lfMenuFont.lfFaceName, &fsize );
					/* */
					RegCloseKey ( hFontInfoKey );
				}
				hMenuFont = CreateFontIndirect ( &lfMenuFont );
				/* create menu */
				hPopup = CreatePopupMenu ( );
				/* fill */
				for ( dummy = 0; dummy < wndlistcnt; ++dummy )
				{
					AppendMenu ( hPopup, MF_BYCOMMAND | MF_ENABLED | MF_OWNERDRAW, dummy + 1, ( wchar_t * ) DCISIG );
				}
				AppendMenu ( hPopup, MF_SEPARATOR, 0, NULL );
				AppendMenu ( hPopup, MF_BYCOMMAND | MF_ENABLED | MF_STRING, DCISIG, L"About..." );
				/* prepare for menu show */
				g_bHookRightPressed = FALSE;
				g_iHookDigitPressed = 0;
				g_hHook = fpSetWindowsHookEx ( WH_KEYBOARD_LL, HookProc, NULL, 0 );
				dummy = GetMessagePos ( );
				/* show */
				imenu = TrackPopupMenuEx ( hPopup, TPM_RIGHTALIGN | TPM_TOPALIGN | TPM_RETURNCMD, rtBar.right, rtBar.bottom, hWnd, NULL );
				/* remove keyboard hook */
				if ( g_iHookDigitPressed )
				{
					imenu = ( g_iHookDigitPressed == 10 ) ? DCISIG : g_iHookDigitPressed;
				}
				if ( g_hHook ) fpUnhookWindowsHookEx ( g_hHook );
				/* parse results */
				if ( imenu )
				{
					if ( imenu <= wndlistcnt )
					{
						DWORD npos = GetMessagePos ( );
						if ( 
							( g_bHookRightPressed || ( ( dummy != npos ) && ( /* x */ LOWORD( npos ) > ( rtBar.right - rtBar.bottom ) ) && ( /* y */ HIWORD( npos ) > rtBar.bottom ) ) )
							&& CanCloseWindow ( wndlist[imenu - 1].hWnd ) )
						{
							DestroyWindow ( wndlist[imenu - 1].hWnd );
						}
						else
						{
							SetForegroundWindow ( wndlist[imenu - 1].hWnd );
						}
					}
					else if ( imenu == DCISIG )
					{
						MessageBox ( hWnd, L" 2007 dci@cpp.in", L"dciTask 0.2a", MB_TOPMOST | MB_SETFOREGROUND );
					}
				}
				/* free menu */
				DestroyMenu ( hPopup );
				/* free menu font */
				DeleteObject ( hMenuFont );
				/* free windows list */
				while ( wndlistcnt-- )
				{
					free ( wndlist[wndlistcnt].sText );
					if ( wndlist[wndlistcnt].hIcon && ( FALSE == wndlist[wndlistcnt].bStockIcon ) ) DestroyIcon ( wndlist[wndlistcnt].hIcon );
				}
				free ( wndlist );
			}
			return 0;
		}
	}
	else if ( nMsg == WM_MEASUREITEM )
	{
		MEASUREITEMSTRUCT * lpmis = ( MEASUREITEMSTRUCT * ) lParam;
		/* */
		if ( wndlistcnt && ( lpmis->CtlType == ODT_MENU ) && ( lpmis->itemID <= wndlistcnt ) && ( lpmis->itemData == DCISIG ) )
		{
			HDC tdc = GetDC ( hWnd );
			/* calc text size */
			SIZE tsize;
			HFONT hNativeFont = ( HFONT ) SelectObject ( tdc, hMenuFont );
			GetTextExtentPoint ( tdc, wndlist[lpmis->itemID - 1].sText, wcslen ( wndlist[lpmis->itemID - 1].sText ), &tsize );
			SelectObject ( tdc, hNativeFont );
			/* write out sizes */
			lpmis->itemHeight = ( GetDeviceCaps ( tdc, LOGPIXELSY ) * 3 ) / 16;
			lpmis->itemWidth  = tsize.cx + 3 * lpmis->itemHeight;
			/* */
			ReleaseDC ( hWnd, tdc );
			return TRUE;
		}
	}
	else if ( nMsg == WM_DRAWITEM )
	{
		DRAWITEMSTRUCT * lpdis = ( DRAWITEMSTRUCT * ) lParam;
		/* */
		if ( wndlistcnt && ( lpdis->CtlType == ODT_MENU ) && ( lpdis->itemID <= wndlistcnt ) && ( lpdis->itemData == DCISIG ) )
		{
			HFONT    hNativeFont;
			int      iNativeBkMode;
			COLORREF cNativeTextColor;
			SIZE tsize;
			RECT rtClip;
			/* calc sizes & offsets */
			int isize = GetDeviceCaps ( lpdis->hDC, LOGPIXELSX ) / 6;
			int ioffs = ( ( lpdis->rcItem.bottom - lpdis->rcItem.top ) - isize ) / 2;
			int bsize = isize - 4 * ioffs;
			int bxoff = ioffs * 4;
			int byoff = ioffs * 2;
			/* erase background */
			FillRect ( lpdis->hDC, &lpdis->rcItem, GetSysColorBrush ( ( lpdis->itemState & ODS_SELECTED ) ? COLOR_HIGHLIGHT : COLOR_MENU ) );
			/* draw icon */
			if ( wndlist[lpdis->itemID - 1].hIcon )
			{
				DrawIconEx ( lpdis->hDC, lpdis->rcItem.left + ioffs, lpdis->rcItem.top + ioffs, wndlist[lpdis->itemID - 1].hIcon, isize, isize, 0, NULL, DI_NORMAL );
			}
			/* draw text */
			hNativeFont      = ( HFONT ) SelectObject ( lpdis->hDC, hMenuFont );
			iNativeBkMode    = SetBkMode ( lpdis->hDC, TRANSPARENT );
			cNativeTextColor = SetTextColor ( lpdis->hDC, GetSysColor ( ( lpdis->itemState & ODS_SELECTED ) ? COLOR_HIGHLIGHTTEXT : COLOR_MENUTEXT ) );
			/* */
			GetTextExtentPoint ( lpdis->hDC, wndlist[lpdis->itemID - 1].sText, wcslen ( wndlist[lpdis->itemID - 1].sText ), &tsize );
			rtClip.left   = lpdis->rcItem.left + isize + 3 * ioffs;
			rtClip.top    = lpdis->rcItem.top + ( ( lpdis->rcItem.bottom - lpdis->rcItem.top ) - tsize.cy ) / 2;
			rtClip.right  = lpdis->rcItem.right - bsize - 2 * bxoff;
			rtClip.bottom = lpdis->rcItem.bottom;
			ExtTextOut ( lpdis->hDC, rtClip.left, rtClip.top, ETO_CLIPPED, &rtClip, wndlist[lpdis->itemID - 1].sText, wcslen ( wndlist[lpdis->itemID - 1].sText ), NULL ); 
			/* */
			SetTextColor ( lpdis->hDC, cNativeTextColor );
			SetBkMode ( lpdis->hDC, iNativeBkMode );
			SelectObject ( lpdis->hDC, hNativeFont );
			/* draw close button */
			if ( CanCloseWindow ( wndlist[lpdis->itemID - 1].hWnd ) )
			{
				POINT ptCloseButton[] = { { lpdis->rcItem.right - bxoff - bsize, lpdis->rcItem.top + byoff }, { lpdis->rcItem.right - bxoff, lpdis->rcItem.top + byoff }, { lpdis->rcItem.right - bxoff, lpdis->rcItem.top + byoff + bsize }, { lpdis->rcItem.right - bxoff - bsize, lpdis->rcItem.top + byoff + bsize }, { lpdis->rcItem.right - bxoff - bsize, lpdis->rcItem.top + byoff }, { lpdis->rcItem.right - bxoff, lpdis->rcItem.top + byoff + bsize }, { lpdis->rcItem.right - bxoff - bsize, lpdis->rcItem.top + byoff + bsize }, { lpdis->rcItem.right - bxoff, lpdis->rcItem.top + byoff } };
				HPEN hNativePen = ( HPEN ) SelectObject ( lpdis->hDC, CreatePen ( PS_SOLID, 1, GetSysColor ( ( lpdis->itemState & ODS_SELECTED ) ? COLOR_HIGHLIGHTTEXT : COLOR_MENUTEXT ) ) );
				Polyline ( lpdis->hDC, ptCloseButton, sizeof ( ptCloseButton ) / sizeof ( * ptCloseButton ) );
				DeleteObject ( SelectObject ( lpdis->hDC, hNativePen ) );
			}
			/* */
			return TRUE;
		}
	}
	else if ( nMsg == WM_PAINT )
	{
		RECT rtBar;
		POINT pt[2];
		LRESULT ret = CallWindowProc ( ( WNDPROC ) lOrigProc, hWnd, nMsg, wParam, lParam );
		/* */
		HDC dcBar = GetDC ( hWnd );
		HPEN hPen = CreatePen ( PS_SOLID, 1, GetSysColor ( COLOR_HIGHLIGHT ) );
		hPen = ( HPEN ) SelectObject ( dcBar, hPen );
		/* */
		GetClientRect ( hWnd, &rtBar );
		/* */
		pt[0].x = rtBar.right - ( rtBar.bottom / 3 );
		pt[0].y = rtBar.bottom / 4;
		pt[1].x = rtBar.right - ( rtBar.bottom / 3 );
		pt[1].y = 3 * ( rtBar.bottom / 4 );
		Polyline ( dcBar, pt, 2 );
		/* */
		hPen = ( HPEN ) SelectObject ( dcBar, hPen );
		DeleteObject ( hPen );
		/* */
		hPen = CreatePen ( PS_SOLID, 1, GetSysColor ( COLOR_HIGHLIGHTTEXT ) );
		hPen = ( HPEN ) SelectObject ( dcBar, hPen );
		/* */
		pt[0].x = ++pt[1].x;
		Polyline ( dcBar, pt, 2 );
		/* */
		hPen = ( HPEN ) SelectObject ( dcBar, hPen );
		DeleteObject ( hPen );
		ReleaseDC ( hWnd, dcBar );
		/* */
		return ret;
	}
	/* */
	return CallWindowProc ( ( WNDPROC ) lOrigProc, hWnd, nMsg, wParam, lParam );
}

int WinMain ( HINSTANCE hInst, HINSTANCE hPrevInst, LPWSTR lpCmdLine, int nCmdShow )
{
	HANDLE hExitEvent = CreateEvent ( NULL, FALSE, FALSE, L"dciTaskExitEvent" );
	/* */
	if ( ERROR_ALREADY_EXISTS == GetLastError ( ) )
	{
		SetEvent ( hExitEvent );
	}
	else
	{
		HWND hBar = FindWindow ( L"HHTaskBar", NULL );
		if ( hBar )
		{
			/* init hooking */
			HMODULE l_coredll     = GetModuleHandle ( L"coredll.dll" );
			fpSetWindowsHookEx    = ( SETWHOOK ) GetProcAddress ( l_coredll, L"SetWindowsHookExW" );
			fpUnhookWindowsHookEx = ( UNHOOKWH ) GetProcAddress ( l_coredll, L"UnhookWindowsHookEx" );
			fpCallNextHookEx      = ( CALLNEXT ) GetProcAddress ( l_coredll, L"CallNextHookEx" );
			/* hook bar */
			lOrigProc  = SetWindowLong ( hBar, GWL_WNDPROC, ( LONG ) FilterProc );
			InvalidateRect ( hBar, NULL, TRUE );
			while ( WAIT_TIMEOUT == WaitForSingleObject ( hExitEvent, INFINITE ) );
			SetWindowLong ( hBar, GWL_WNDPROC, lOrigProc );
			InvalidateRect ( hBar, NULL, TRUE );		
		}
		else
		{
			MessageBox ( NULL, L"  !", L"dciTask", MB_TOPMOST | MB_ICONEXCLAMATION );
		}
	}
	/* */
	CloseHandle ( hExitEvent );
}
